from django.db.models import CharField

from .form_fields import StaticFormField

__all__ = ['StaticField']


class StaticField(CharField):

    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = 20
        kwargs['default'] = ''
        kwargs['blank'] = True
        kwargs['null'] = True
        kwargs['name'] = ''
        kwargs['verbose_name'] = ''
        super().__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        return super().formfield(
            **{
                "form_class": StaticFormField,
                **kwargs,
            }
        )



import logging
import inspect

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.db.models.signals import pre_save, post_init, post_save, pre_delete
from django.dispatch import receiver

from django.db import models as django_models
from django.core.validators import MinValueValidator

from django.utils.translation import gettext_lazy as _
from django.utils import timezone
from django.utils.functional import cached_property, classproperty
from django.conf import settings

from taggit.models import TagBase, ItemBase

from modelcluster.contrib.taggit import ClusterTaggableManager
from modelcluster.models import ClusterableModel
from modelcluster.fields import ParentalKey

from wagtail.models import Page, Orderable
from wagtail.fields import RichTextField
from wagtail.admin.panels import FieldPanel, MultiFieldPanel, PageChooserPanel, ObjectList, InlinePanel
from wagtail.snippets.models import register_snippet

from django_auxiliaries.model_fields import MultipleChoiceField
from wagtail_block_model_field.fields import BlockModelField

from officekit import models as officekit # noqa

from .blocks import *  # We import blocks here so that the content block registration methods are called.
from .validators import validate_identifier, validate_category
from .mixins import SpecificMixin
from .fields import StaticField
from .apps import get_app_label

__all__ = ['Synopsis', 'PageSynopsis', 'build_synopsis_class_for', 'SynopsisAdapter',
           'sync_page_synopsis', 'determine_page_url', 'SynopsisTagItem', 'SynopsisTag']

APP_LABEL = get_app_label()

logger = logging.getLogger("passenger_wsgi")


class SynopsisTag(TagBase):
    class Meta:
        verbose_name = "synopsis tag"
        verbose_name_plural = "synopsis tags"


class SynopsisTagItem(ItemBase):

    class Meta:
        # constraints = [
        #    django_models.UniqueConstraint('content_object',
        #                                   'tag', name='unique_%(app_label)s_%(class)s.unique_tag_assignment')
        #]

        unique_together = (('content_object', 'tag'),)

    tag = django_models.ForeignKey(
        SynopsisTag, related_name="tagged_synopsis", on_delete=django_models.CASCADE
    )

    content_object = ParentalKey(
        to=APP_LABEL + '.Synopsis',
        on_delete=django_models.CASCADE,
        related_name='tagged_items'
    )


AUTHOR_ROLE = 'author'
EDITOR_ROLE = 'editor'

ROLES = (AUTHOR_ROLE, EDITOR_ROLE)
ROLE_CHOICES = ((AUTHOR_ROLE, "Author"), (EDITOR_ROLE, "Editor"))


def get_default_synopsis_content_type():
    """
    Returns the content type to use as a default for media items whose content type
    has been deleted.
    """
    return ContentType.objects.get_for_model(Synopsis)


# noinspection SpellCheckingInspection
@register_snippet
class Synopsis(SpecificMixin, Orderable, ClusterableModel):

    class Meta:
        verbose_name = 'synopsis'
        verbose_name_plural = 'synopses'

    AUTHOR_NAMES = 'author'
    EDITOR_NAMES = 'editor'

    NAMES_ROLES = [AUTHOR_NAMES, EDITOR_NAMES]

    @property
    def last_date_time(self):
        return self.updated_at if self.updated_at else self.created_at

    @property
    def timestamp_category(self):
        if self.updated_at == self.created_at:
            return "created"

        return "updated"

    @property
    def represents(self):

        if self.specific != self:
            return self.specific.represents

        return None

    @property
    def has_destination_url(self):

        if self.specific != self:
            url = self.specific.determine_destination_url()
        else:
            url = self.determine_destination_url()

        return bool(url)

    static = StaticField()
    # This is a placeholder so that the default unchanged form is accepted as an instance on post

    content_type = django_models.ForeignKey(
        ContentType,
        verbose_name=_('content type'),
        related_name='synopses',
        on_delete=django_models.SET(get_default_synopsis_content_type),
        editable=False
    )

    category = django_models.CharField(max_length=128, validators=[validate_identifier], default='')

    rubric = django_models.CharField(max_length=128, default='', blank=True)
    rubric_url = django_models.URLField(verbose_name=_("rubric URL"), default='', blank=True)

    heading = django_models.CharField(max_length=512, default='', blank=True)
    summary = RichTextField(default='', blank=True, editor=APP_LABEL + ".summary")

    visual = BlockModelField(VisualBlock(required=False), value_class=VisualBlockValue, blank=True, null=True)

    names = BlockModelField(NameListBlock(choices=ROLE_CHOICES, default_choice=AUTHOR_ROLE),
                            value_class=NameListBlockValue, blank=True, null=True)

    external_url = django_models.URLField(
        verbose_name=_("external URL"),
        default='',
        blank=True
    )

    tags = ClusterTaggableManager(through=APP_LABEL + '.SynopsisTagItem', blank=True)

    prominence_index = django_models.IntegerField(default=None,
                                                  validators=[
                                                        MinValueValidator(1)
                                                  ],
                                                  blank=True,
                                                  null=True)

    prominence_order = django_models.IntegerField(default=None,
                                                  validators=[
                                                        MinValueValidator(1)
                                                  ],
                                                  blank=True,
                                                  null=True)

    created_by_user = django_models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_('created by user'),
        null=True,
        blank=True,
        editable=False,
        on_delete=django_models.SET_NULL
    )

    created_at = django_models.DateTimeField(
        verbose_name=_('created at'),
        default=None,
        editable=False)

    updated_at = django_models.DateTimeField(
        verbose_name=_('updated at'),
        blank=True,
        null=True,
        default=None,
        editable=False)

    metrics = django_models.JSONField(blank=True, null=True, default=dict)

    panels = [
        FieldPanel('category'),

        MultiFieldPanel([
            FieldPanel('heading'),
            FieldPanel('summary'),
            FieldPanel('tags')
        ], heading='Brief'
        ),

        FieldPanel('visual'),
        InlinePanel('name_index', heading="Indexed Names"),
        MultiFieldPanel([
            FieldPanel('external_url'),
        ], heading='Destination URL'
        ),

        MultiFieldPanel([
            FieldPanel('prominence_index'),
            FieldPanel('prominence_order'),
        ], heading='Prominence'
        )
    ]

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        # noinspection PyUnresolvedReferences
        if not self.id:

            self.id = None
            self.created_at = timezone.now()
            self.updated_at = self.created_at

            # this model is being newly created
            # rather than retrieved from the db;
            # noinspection PyUnresolvedReferences
            if not self.content_type_id:

                # set content type to correctly represent the model class
                # that this was created as
                self.content_type_id = ContentType.objects.get_for_model(self).id

        self.deferred_name_assignments = []

    def determine_destination_url(self, request=None, current_site=None):

        if self.external_url:
            return self.external_url

        if not self.represents:
            return ''

        result = ''

        for adapter in self.specific_class.adapters:
            determine_url_method = adapter.get_determine_url_method(self.represents)

            if not determine_url_method:
                continue

            result = determine_url_method(request, current_site)
            break

        return result

    def save(self, **kwargs):

        for adapter in self.specific_class.adapters:
            sync_method = adapter.get_sync_method(self.represents)

            if not sync_method:
                continue

            sync_method(self)

        self.deferred_name_assignments = []

        json_value = self.specific_class.names.field.block_def.get_prep_value(self.names) # noqa

        role = json_value.get('role', None) # noqa

        self.index_names_from(json_value, role)

        super().save(**kwargs)

        self.clean_name_index(role)

        for assignment in self.deferred_name_assignments:
            assignment.synopsis = self.specific
            assignment.save()

        self.deferred_name_assignments = []

    @classproperty
    def adapters(cls):
        return cls._adapters

    @classmethod
    def install_synopsis_adapter_for(cls, editable_fields=None, sync_method=None, determine_url_method=None,
                                     tagged_item_class=None, tags_field='tags', tagged_items_field='tagged_items'):

        model_adapter = SynopsisAdapter()
        model_adapter.synopsis_class = cls
        model_adapter.editable_fields = editable_fields
        model_adapter.sync_method = sync_method
        model_adapter.determine_url_method = determine_url_method
        model_adapter.tagged_item_class = tagged_item_class
        model_adapter.tags_field = tags_field
        model_adapter.tagged_items_field = tagged_items_field

        cls.adapters.append(model_adapter)

        model_adapter.install_signal_handlers()

        return model_adapter

    @cached_property
    def name_assignment_model_class(self):
        return apps.get_model(APP_LABEL + ".nameinsynopsis")

    def create_name_assignment(self, family_name, given_names_and_initials, person, role):

        assignment = self.name_assignment_model_class(
                            synopsis_id=self.pk,
                            family_name=family_name,
                            given_names_and_initials=given_names_and_initials,
                            person_id=person.pk,
                            role=role)

        self.deferred_name_assignments.append(assignment)

    def clean_name_index(self, role):
        filters = {
            'synopsis_id': self.pk
        }

        if role:
            filters['role'] = role

        self.name_assignment_model_class.objects.filter(**filters).delete()

    def index_names_from(self, value, role):

        if value is None:
            return

        names = value.get("names", [])
        names = [name for name in names if name["type"] == "person"]

        for idx, name in enumerate(names):

            name_type = name['type']

            if name_type != 'person':
                continue

            person_name = name['value']

            given_names_and_initials = person_name['given_names_and_initials']
            family_name = person_name['family_name']

            given_names_and_initials = given_names_and_initials.strip()
            family_name = family_name.strip()

            identify_with = person_name.get("identify_with", None)

            person = None

            if identify_with:

                try:
                    person = officekit.Person.objects.get(pk=identify_with)
                except officekit.Person.DoesNotExist:
                    pass
            else:

                people = officekit.Person.objects.all().filter(
                    given_names_and_initials__iexact=given_names_and_initials,
                    family_name__iexact=family_name)

                n = people.count()

                if n == 1:
                    person = people[0]

            if person:

                given_names_and_initials = person.given_names_and_initials
                family_name = person.family_name

                given_names_and_initials = given_names_and_initials.strip()
                family_name = family_name.strip()

            self.create_name_assignment(family_name=family_name, given_names_and_initials=given_names_and_initials,
                                        person=person, role=role)

    def __str__(self):

        result = ''

        if self.category:
            result += '[{}] '.format(self.category)

        text = self.heading

        if not text:
            text = self.summary

        result += text
        result += " ▶ " + self.determine_destination_url()
        return result


@register_snippet
class NameInSynopsis(Orderable):

    class Meta(Orderable.Meta):
        verbose_name = 'Name in Synopsis'
        verbose_name_plural = 'Names in Synopsis'
        # constraints = [
        #    django_models.UniqueConstraint(fields=['synopsis', 'family_name', 'given_names_and_initials'],
        #                                   name='unique_%(app_label)s_%(class)s.name_assignment')
        #]

        unique_together = (('synopsis', 'family_name', 'given_names_and_initials'),)

    synopsis = ParentalKey(Synopsis, related_name="name_index", on_delete=django_models.CASCADE, blank=False)

    family_name = django_models.CharField("Family Name", default="", max_length=192, blank=False)
    given_names_and_initials = django_models.CharField("Given Name(s) and Initial(s)", default="", max_length=192, blank=False)

    person = django_models.ForeignKey(officekit.Person, related_name="synopses+",
                                      on_delete=django_models.SET_NULL, blank=True, null=True)

    role = MultipleChoiceField(choices=ROLE_CHOICES, default=AUTHOR_ROLE, blank=True, max_length=64)

    panels = [
        FieldPanel('synopsis'),
        FieldPanel('family_name'),
        FieldPanel('given_names_and_initials'),
        FieldPanel('person'),
        FieldPanel('role'),
    ]

    @classmethod
    def create(cls, synopsis, personal_name, person, role):

        result = cls(synopsis=synopsis, personal_name=personal_name, person=person, role=role)

        return result

    def __str__(self):
        return self.given_names_and_initials + " " + self.family_name + " in "  + self.synopsis.heading + (" [identified]" if self.person else "") # noqa


class SynopsisAdapter:

    synopsis_class = Synopsis
    editable_fields = None

    sync_method = None
    sync_method_name = 'sync_synopsis'

    determine_url_method = None
    determine_url_method_name = 'determine_url'

    tagged_item_class = None
    tags_field = None
    tagged_items_field = None

    @cached_property
    def synopsis_content_type(self):
        return ContentType.objects.get_for_model(self.synopsis_class)

    def create_synopsis(self):
        return self.synopsis_class()

    def get_synopsis_field_name(self, instance_or_class):
        synopsis_field_name = self.synopsis_class.model.field.remote_field.name
        synopsis_field_name = getattr(instance_or_class, 'synopsis_field_name', synopsis_field_name)
        return synopsis_field_name

    def get_synopsis_editable_fields(self, instance_or_class):
        synopsis_editable_fields = getattr(instance_or_class, 'synopsis_editable_fields', self.editable_fields)

        if synopsis_editable_fields is None:
            synopsis_editable_fields = list(self.synopsis_class._meta.fields)

        return synopsis_editable_fields

    def get_sync_method(self, instance):
        sync_method = getattr(instance, self.sync_method_name, None)

        if sync_method is None and self.sync_method:

            def sync_method(synopsis):
                return self.sync_method(instance, synopsis)

        if not callable(sync_method):
            return None

        return sync_method

    def get_determine_url_method(self, instance):
        determine_url_method = getattr(instance, self.determine_url_method_name, None)

        if determine_url_method is None and self.determine_url_method:

            def determine_url_method(request, current_site=None):
                return self.determine_url_method(instance, request, current_site=current_site)

        if not callable(determine_url_method):
            return None

        return determine_url_method

    def get_synopsis_relation(self, instance):
        synopsis_field_name = self.get_synopsis_field_name(instance)
        synopsis_relation = getattr(instance, synopsis_field_name, None)
        return synopsis_relation

    def create_synopsis_panel_for(self, model_class):

        synopsis_field_name = self.get_synopsis_field_name(model_class)
        synopsis_editable_fields = self.get_synopsis_editable_fields(model_class)

        synopsis_panel = ObjectList([
            InlinePanel(synopsis_field_name, min_num=1, max_num=1, panels=[
                             FieldPanel('static'),
                         ] + [FieldPanel(field) for field in
                              synopsis_editable_fields if field != 'static']
            )
        ],
            heading='Synopsis'
        )

        synopsis_panel = synopsis_panel.bind_to_model(model_class)

        return synopsis_panel

    # noinspection PyMethodMayBeStatic
    def tag_pre_delete(self, instance, **kwargs):

        if not instance.content_object.synopsis.exists():
            return

        try:
            shadow_tag = SynopsisTag.objects.get(name=instance.tag.name, slug=instance.tag.slug) # noqa
        except SynopsisTag.DoesNotExist: # noqa
            return

        for synopsis in instance.content_object.synopsis.all():

            shadow_items = SynopsisTagItem.objects.filter(content_object_id=synopsis.id, tag_id=shadow_tag.id) # noqa
            shadow_items.delete()

    # noinspection PyMethodMayBeStatic
    def synopsis_post_save(self, instance, **kwargs):

        for tag in getattr(instance.represents, self.tags_field).all():
            shadow_tag, _ = SynopsisTag.objects.get_or_create(name=tag.name, slug=tag.slug) # noqa

            shadow_item, created = SynopsisTagItem.objects.get_or_create( # noqa
                content_object_id=instance.id, tag_id=shadow_tag.id)

            if not created:
                continue

            getattr(instance, self.tagged_items_field).add(shadow_item)
            # print("Saving tag \'%s\' for \'%s\'" % (shadow_tag.name, instance.heading))

    def install_signal_handlers(self):

        if self.tagged_item_class is None or not self.tags_field or not self.tagged_items_field:
            return

        def synopsis_post_save_wrapper(instance, **kwargs):
            self.synopsis_post_save(instance, **kwargs)

        def tag_pre_delete_wrapper(instance, **kwargs):
            self.tag_pre_delete(instance, **kwargs)

        receiver(post_save, sender=self.synopsis_class)(synopsis_post_save_wrapper)
        receiver(pre_delete, sender=self.tagged_item_class)(tag_pre_delete_wrapper)


def build_synopsis_class_for(model_class,
                             field_on_model=None):

    if not isinstance(model_class, ClusterableModel):
        RuntimeError('Model class for Synopsis must be derived from ClusterableModel')

    name = model_class.__name__ + 'Synopsis'

    if not field_on_model:
        field_on_model = 'synopsis'

    stack = inspect.stack()[1]
    module = inspect.getmodule(stack[0])

    def represents(self):
        return self.model

    synopsis_class = type(name, (Synopsis,), {
        '__module__': module.__name__,
        'model': ParentalKey(model_class, related_name=field_on_model, on_delete=django_models.CASCADE, blank=False),
        'model_class': model_class,
        'represents': property(fget=represents),
        '_adapters': []
    })

    return synopsis_class

"""
def install_signal_handlers(model_class, model_adapter, exclude_models=None):

    if exclude_models is None:
        exclude_models = []

    stack = [(model_class, 0, None)]

    while stack:

        model, child_index, children = stack[-1]

        if children is None:

            children = list(model.__subclasses__())

            if model not in exclude_models:

                # print(model.__name__)

                decorator = receiver(post_init, sender=model)
                decorator(lambda signal, **kwargs: model_adapter.model_post_init(**kwargs))

                decorator = receiver(pre_save, sender=model)
                decorator(lambda signal, **kwargs: model_adapter.model_pre_save(**kwargs))

        if child_index >= len(children):
            stack.pop()
            continue

        stack[-1] = model, child_index + 1, children
        stack.append((children[child_index], 0, None))

    pass
"""


def sync_page_synopsis(page, synopsis):

    synopsis.category = getattr(page, 'synopsis_category', '')
    synopsis.heading = page.title
    synopsis.rubric = ''
    synopsis.rubric_url = ''

    synopsis.created_by_user = page.owner
    synopsis.created_at = page.first_published_at or timezone.now()
    synopsis.updated_at = page.last_published_at

    if not synopsis.summary:
        pass


def determine_page_url(page, request, current_site=None):
    return page.get_url(request=request, current_site=current_site)


PageSynopsis = build_synopsis_class_for(Page)



# Generated by Django 4.1 on 2023-03-07 19:20

from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django_auxiliaries.model_fields
import modelcluster.contrib.taggit
import modelcluster.fields
import wagtail.blocks.struct_block
import wagtail.fields
import wagtail_block_model_field.fields
import wagtail_switch_block.blocks
import wagtail_synopsis.blocks
import wagtail_synopsis.fields
import wagtail_synopsis.mixins
import wagtail_synopsis.models
import wagtail_synopsis.validators


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wagtailcore', '0078_referenceindex'),
        ('contenttypes', '0002_remove_content_type_name'),
        ('officekit', '0002_add_address_formats'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Synopsis',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sort_order', models.IntegerField(blank=True, editable=False, null=True)),
                ('static', wagtail_synopsis.fields.StaticField(blank=True, default='', max_length=20, null=True, verbose_name='')),
                ('category', models.CharField(default='', max_length=128, validators=[wagtail_synopsis.validators.validate_identifier])),
                ('rubric', models.CharField(blank=True, default='', max_length=128)),
                ('rubric_url', models.URLField(blank=True, default='', verbose_name='rubric URL')),
                ('heading', models.CharField(blank=True, default='', max_length=512)),
                ('summary', wagtail.fields.RichTextField(blank=True, default='')),
                ('visual', wagtail_block_model_field.fields.BlockModelField(wagtail_synopsis.blocks.VisualBlock(required=False), wagtail_switch_block.blocks.SwitchValue)),
                ('names', wagtail_block_model_field.fields.BlockModelField(wagtail_synopsis.blocks.NameListBlock(choices=(('author', 'Author'), ('editor', 'Editor')), default_choice='author'), wagtail.blocks.struct_block.StructValue)),
                ('external_url', models.URLField(blank=True, default='', verbose_name='external URL')),
                ('prominence_index', models.IntegerField(blank=True, default=None, null=True, validators=[django.core.validators.MinValueValidator(1)])),
                ('prominence_order', models.IntegerField(blank=True, default=None, null=True, validators=[django.core.validators.MinValueValidator(1)])),
                ('created_at', models.DateTimeField(default=None, editable=False, verbose_name='created at')),
                ('updated_at', models.DateTimeField(blank=True, default=None, editable=False, null=True, verbose_name='updated at')),
                ('metrics', models.JSONField(blank=True, default=dict, null=True)),
                ('content_type', models.ForeignKey(editable=False, on_delete=models.SET(wagtail_synopsis.models.get_default_synopsis_content_type), related_name='synopses', to='contenttypes.contenttype', verbose_name='content type')),
                ('created_by_user', models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL, verbose_name='created by user')),
            ],
            options={
                'verbose_name': 'synopsis',
                'verbose_name_plural': 'synopses',
            },
            bases=(wagtail_synopsis.mixins.SpecificMixin, models.Model),
        ),
        migrations.CreateModel(
            name='SynopsisTag',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, unique=True, verbose_name='name')),
                ('slug', models.SlugField(allow_unicode=True, max_length=100, unique=True, verbose_name='slug')),
            ],
            options={
                'verbose_name': 'synopsis tag',
                'verbose_name_plural': 'synopsis tags',
            },
        ),
        migrations.CreateModel(
            name='SynopsisTagItem',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('content_object', modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='tagged_items', to='wagtail_synopsis.synopsis')),
                ('tag', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='tagged_synopsis', to='wagtail_synopsis.synopsistag')),
            ],
            options={
                'unique_together': {('content_object', 'tag')},
            },
        ),
        migrations.AddField(
            model_name='synopsis',
            name='tags',
            field=modelcluster.contrib.taggit.ClusterTaggableManager(blank=True, help_text='A comma-separated list of tags.', through='wagtail_synopsis.SynopsisTagItem', to='wagtail_synopsis.SynopsisTag', verbose_name='Tags'),
        ),
        migrations.CreateModel(
            name='PageSynopsis',
            fields=[
                ('synopsis_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtail_synopsis.synopsis')),
                ('model', modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='synopsis', to='wagtailcore.page')),
            ],
            options={
                'ordering': ['sort_order'],
                'abstract': False,
            },
            bases=('wagtail_synopsis.synopsis',),
        ),
        migrations.CreateModel(
            name='NameInSynopsis',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sort_order', models.IntegerField(blank=True, editable=False, null=True)),
                ('family_name', models.CharField(default='', max_length=192, verbose_name='Family Name')),
                ('given_names_and_initials', models.CharField(default='', max_length=192, verbose_name='Given Name(s) and Initial(s)')),
                ('role', django_auxiliaries.model_fields.MultipleChoiceField(blank=True, choices=[('author', 'Author'), ('editor', 'Editor')], default='author', max_length=64)),
                ('person', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='synopses+', to='officekit.person')),
                ('synopsis', modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='name_index', to='wagtail_synopsis.synopsis')),
            ],
            options={
                'verbose_name': 'Name in Synopsis',
                'verbose_name_plural': 'Names in Synopsis',
                'ordering': ['sort_order'],
                'abstract': False,
                'unique_together': {('synopsis', 'family_name', 'given_names_and_initials')},
            },
        ),
    ]




from django.db import transaction


def replicate_tags(tags, tag_class):
    tag_names = [tag.name for tag in tags]
    result = []

    with transaction.atomic():
        for tag_name in tag_names:
            try:
                replicated_tag = tag_class.objects.all().get(name=tag_name)
            except tag_class.DoesNotExist:
                replicated_tag = tag_class(name=tag_name)
                replicated_tag.save()

            result.append(replicated_tag)

    return result


def apply_synopsis_tags_to_model(synopsis_tags, model, model_tag_class, model_tag_item_class):
    model_tags = replicate_tags(synopsis_tags.all(), model_tag_class)
    model_items = []

    for model_tag in model_tags:
        model_item = model_tag_item_class(tag=model_tag, content_object=model)
        model_items.append(model_item)

    model.tagged_items = model_items




import re

from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ValidationError

__all__ = ['validate_identifier', 'validate_category']


def validate_identifier(value):

    if re.search(r"(^[^A-Za-z0-9_]|\s)", value):

        raise ValidationError(
            _("Identifiers must start with a letter, a digit or underscore "
              "and cannot contain whitespace characters: {:}").format(value)
        )


def validate_category(value):

    if re.search(r"(^[^A-Za-z0-9_]|\s)", value):

        raise ValidationError(
            _("Categories must start with a letter, a digit or underscore "
              "and cannot contain whitespace characters: {:}").format(value)
        )



WAGTAIL_SYNOPSIS_SYNOPSES_QUERY_FORM_CLASSNAME = ''
WAGTAIL_SYNOPSIS_SYNOPSES_LAYOUT_BLOCK_CLASSNAME = 'curated-content'
WAGTAIL_SYNOPSIS_SYNOPSIS_LIST_BLOCK_CLASSNAME = 'curated-content'
WAGTAIL_SYNOPSIS_SYNOPSIS_LIST_CONTENT_WRAPPER_CLASSNAME = ''

WAGTAIL_SYNOPSIS_TAG_CLASSNAME = 'synopsis-tag'
WAGTAIL_SYNOPSIS_NAMES_CLASSNAME = 'abstract-detail abstract-byline'




from model_porter.config import ModelPorterConfig
from model_porter.utilities import define_tags as define_generic_tags
from model_porter.repository import UndefinedReference

from .models import SynopsisTag, SynopsisTagItem


def define_tags(*, tag_values, context):
    return define_generic_tags(tag_values=tag_values, tag_class=SynopsisTag, tag_item_class=SynopsisTagItem, context=context)


def create_synopsis(*, model, context, for_object):

    for_object_instance = context.get_instance(for_object, None)

    if for_object_instance is None:
        raise UndefinedReference(for_object)

    try:
        result = for_object_instance.synopsis.get()
    except model.DoesNotExist:
        result = model()
        for_object_instance.synopsis.add(result)

    return result


class WagtailSynopsisConfig(ModelPorterConfig):

    def __init__(self, app_label, module):
        super(WagtailSynopsisConfig, self).__init__(app_label, module)
        self.register_function_action(create_synopsis, context_argument='context')
        self.register_function_action(define_tags, context_argument='context')




import weakref

from django.conf import settings

from wagtail import blocks

from wagtail_tags_block.blocks import TagsBlock
from wagtail_switch_block import SwitchBlock, DynamicSwitchBlock
from wagtail_switch_block.block_registry import BlockRegistry

from wagtail_dynamic_choice.blocks import DynamicMultipleChoiceBlock

from wagtail_content_block.annotations import *
from wagtail_content_admin.blocks import ContentBlock, ContentBlockValue, register_content_block, ContentProviderBlockMixin

from media_catalogue.blocks import MediaItemChooserBlock, MediaItemChooserBlockValue

from querykit.base import *
from querykit.value_types import *
from querykit.value_sources import *
from querykit.forms import PickerFactory, SelectFactory, Choice

from csskit.blocks import SideBlock
from aldine.blocks import BaseContentLayoutBlock
from figurative.blocks import FigureBlock, SlideshowBlock
from officekit.blocks import NameListBlock as DefaultNameListBlock, NameListBlockValue as DefaultNameListBlockValue

from .apps import get_app_label
from .curation import CuratedSynopsis

__all__ = ['NameListBlock', 'NameListBlockValue',
           'VisualBlock', 'VisualBlockValue',
           'SynopsesBlock', 'SynopsesBlockValue', 'SynopsesLayoutBlock', 'SynopsisListBlock',
           'register_curator_block', # noqa
           'StandardCuratorBlock']

APP_LABEL = get_app_label()

NameListBlockValue = DefaultNameListBlockValue


class NameListBlock(DefaultNameListBlock):

    class Meta:
        classname = settings.WAGTAIL_SYNOPSIS_NAMES_CLASSNAME

    role = blocks.ChoiceBlock(choices=[], default=None, required=False)

    def __init__(self, local_blocks=None, choices=None, default_choice=None, **kwargs):

        if choices is None:
            choices = []

        if local_blocks:
            local_block_dict = dict(local_blocks)
        else:
            local_block_dict = {}

        local_block_dict["role"] = blocks.ChoiceBlock(choices=choices, default=default_choice)

        local_blocks = local_block_dict.items()

        self.constructor_kwargs_ = kwargs

        super().__init__(local_blocks, **kwargs)

        self.constructor_kwargs_["choices"] = choices
        self.constructor_kwargs_["default_choice"] = default_choice

    def deconstruct(self):
        return blocks.Block.deconstruct(self)


VISUAL_ANNOTATIONS = ContentAnnotations(groups=[

            ContentAnnotationGroup(
                identifier='credits',
                label='Credits',
                fields=[
                    TextAnnotationField(
                        identifier='category',
                        label='',
                        default_value='Image',
                        attributes={
                            'placeholder': ''
                        }
                    ),

                    TextAreaAnnotationField(
                        identifier='names',
                        label='Name(s)',
                        attributes={
                            'placeholder': ''
                        }
                    )
                ]
            )
        ])


VisualBlockValue = MediaItemChooserBlockValue


class VisualBlock(MediaItemChooserBlock):

    class Meta:
        annotations = VISUAL_ANNOTATIONS


class CuratorBlockRegistry(BlockRegistry):

    # noinspection PyMethodMayBeStatic
    def should_include_entry_for_switch_block(self, identifier, entry, switch_block):
        return True

    # noinspection PyMethodMayBeStatic
    def instantiate_block(self, identifier, entry, switch_block):

        block_kwargs = dict(entry.block_kwargs)

        block = entry.block_type(*entry.block_args, **block_kwargs)
        block.set_name(identifier)

        return block


CURATOR_BLOCK_REGISTRY = CuratorBlockRegistry()
CURATOR_BLOCK_REGISTRY.define_procedures_in_caller_module("curator")


Synopsis = None


class CuratorBlock(ContentProviderBlockMixin, blocks.StructBlock):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, target_model=APP_LABEL + ".synopsis", **kwargs)

    # noinspection PyMethodMayBeStatic
    def derive_content(self, value, request=None):

        global Synopsis

        if Synopsis is None:
            from .models import Synopsis

        items, _ = self.select(value, Synopsis.objects.all())
        annotations = self.clean_annotations(items)
        return self.create_content(items=items, annotations=annotations)

    # noinspection PyMethodMayBeStatic
    def select(self, curator_value, query_set, request=None):

        global Synopsis

        if Synopsis is None:
            from .models import Synopsis

        return query_set, Synopsis.objects.none()


MAIN_PROMINENCE_INDEX = 1
MAJOR_PROMINENCE_INDEX = 2
MINOR_PROMINENCE_INDEX = 3


class StandardCuratorBlock(CuratorBlock):

    class Meta:
        prominence_css_class_map = {
            MAIN_PROMINENCE_INDEX: 'main',
            MAJOR_PROMINENCE_INDEX: 'major',
            MINOR_PROMINENCE_INDEX: 'minor',
        }
        prominence_slots_map = {
            MAIN_PROMINENCE_INDEX: 2,
            MAJOR_PROMINENCE_INDEX: 2,
            MINOR_PROMINENCE_INDEX: 5,
    }

    categories = DynamicMultipleChoiceBlock(label="Categories", default=[], required=False,
                                            choices_function_name=APP_LABEL + ".curation.get_synopsis_category_choices")

    tags = TagsBlock(tag_model=APP_LABEL + ".synopsistag", required=False, help_text="Select Tagged Synopses Only")

    def select(self, curator_value, query_set, request=None):

        filter_options = {
            'content_object_id__in': query_set.values_list('id', flat=True),
        }

        curator_tag_names = [tag.name for tag in curator_value['tags']]

        if curator_tag_names:
            filter_options['tag__name__in'] = curator_tag_names

            selected_ids = [item.content_object_id for item in SynopsisTagItem.objects.all(). # noqa
                prefetch_related('tag', 'content_object').filter(
                **filter_options
            )]

            selected = query_set.filter(id__in=selected_ids)
        else:
            selected = query_set

        filter_options = {}
        categories = curator_value['categories']

        if categories:
            filter_options['category__in'] = categories

        if filter_options:
            selected = selected.filter(**filter_options)

        selected = selected.order_by('prominence_index', 'prominence_order', '-updated_at')

        prominence_orders = {}
        prominence_css_class_map = dict(self.meta.prominence_css_class_map)
        prominence_slots_map = dict(self.meta.prominence_slots_map)

        curated_selection = []

        for index, item in enumerate(selected):

            curated = CuratedSynopsis(item)

            if not curated.prominence_index:
                curated.prominence_index = 1

            slots = None
            assigned_slot = False

            while slots is None and prominence_slots_map:
                slots = prominence_slots_map.get(curated.prominence_index, None)

                if slots is None:
                    curated.prominence_index += 1
                    continue

                if slots > 0:
                    assigned_slot = True
                    slots -= 1
                    if slots > 0:
                        prominence_slots_map[curated.prominence_index] = slots
                    else:
                        del prominence_slots_map[curated.prominence_index]
                    break

                curated.prominence_index = curated.prominence_index + 1

            if not assigned_slot:
                break

            curated.prominence_order = prominence_orders.setdefault(curated.prominence_index, 0) + 1
            prominence_orders[curated.prominence_index] = curated.prominence_order

            curated.prominence_css_class = prominence_css_class_map.get(curated.prominence_index, '')
            curated_selection.append(curated)

        selected_ids = [curated.id for curated in curated_selection]

        return curated_selection, query_set.exclude(id__in=selected_ids)


register_curator_block(APP_LABEL, "standard", StandardCuratorBlock, [], {}) # noqa


def configure_chooser_synopsis_item(item, instance):

    from .content_admin import synopsis_admin
    synopsis_admin.configure_frontend_item(item, instance)


def create_synopsis_query_parameters():

    order_choices = [OrderSpecifier("date_descending", "↓ Date",
                                    PathValueSource(value_type=DateTimeType, path='-updated_at')
                                    ),

                     OrderSpecifier("date_ascending", "↑ Date",
                                    PathValueSource(value_type=DateTimeType, path='updated_at')
                                    ),

                     OrderSpecifier("title", "↑ Title",
                                    PathValueSource(value_type=TextType, path='heading')
                                    )
                     ]

    select_factory = SelectFactory()

    fp_factory = PickerFactory()

    parameters = [
        ResultSliceParameter(
               identifier='page',
               widget_factory=None),

        OrderParameter(
               identifier='sort_by',
               label='Sort By',
               order_specifiers=order_choices,
               widget_factory=select_factory),

        SliceSizeParameter(
               identifier='per_page',
               label='Per Page',
               size_choices=[
                    Choice(25, "25"),
                    Choice(50, "50"),
                    Choice(100, "100"),
                    Choice('all', "All")
               ],
               widget_factory=select_factory),

        Filter(identifier='author',
               label='Author',
               value_source=PathValueSource(value_type=TextType,
                                            path='name_index.family_name',
                                            label_path='name_index.family_name'),
               widget_factory=fp_factory),

        # Filter(identifier='rubric',
        #       label='Rubric',
        #       value_source=PathValueSource(value_type=TextType, path='rubric'),
        #       widget_factory=fp_factory),

        # Filter(identifier='updated_at',
        #       label='Date',
        #       value_source=PathValueSource(value_type=DateTimeType, path='updated_at'),
        #       widget_factory=fp_factory),

        Filter(identifier='keyword',
               label='Keyword',
               value_source=PathValueSource(value_type=TextType, path='tags.slug', label_path='tags.name'),
               widget_factory=fp_factory)
    ]

    return parameters


def create_synopsis_query_widgets():

    widgets = [
        "clear", "submit"
    ]

    return widgets


SynopsesBlockValue = ContentBlockValue


class SynopsesBlock(ContentBlock):

    class Meta:
        verbose_item_name = "Synopsis"
        verbose_item_name_plural = "Synopses"

        chooser_url = APP_LABEL + ':chooser'
        configure_function_name = APP_LABEL + '.blocks.configure_chooser_synopsis_item'
        max_num_choices = None
        chooser_prompts = {
            'add': 'Add Synopsis',
            'replace': 'Replace Synopsis'
        }

        chooser_filter = None
        content_choice_blocks_function_name = APP_LABEL + '.blocks.curator_block_choices'

        query_field_choices = [
            ('category', 'Category'),
            ('heading', 'Heading'),
            ('summary', 'Summary'),
            ('updated_at', 'Updated Date Time'),
            ('rubric', 'Rubric'),
            ('tags__name', 'Tag Name'),
        ]
        query_slice_size = 25
        query_parameters = create_synopsis_query_parameters()
        query_form_widgets = create_synopsis_query_widgets()
        query_form_classname = settings.WAGTAIL_SYNOPSIS_SYNOPSES_QUERY_FORM_CLASSNAME

    def __init__(self, **kwargs):
        super().__init__(target_model=APP_LABEL + ".synopsis", **kwargs)  # noqa


register_content_block(APP_LABEL, "synopses", SynopsesBlock, [], {}, item_type=APP_LABEL + ".Synopsis")  # noqa


class SynopsisStyleBlock(blocks.StructBlock):

    # synopsis = blocks.StaticBlock()  # placeholder

    def get_context(self, value, parent_context=None):
        context = super(SynopsisStyleBlock, self).get_context(value, parent_context)
        return context


class VisualSynopsisBlock(SynopsisStyleBlock):

    class Meta:
        template = APP_LABEL + "/blocks/visual_synopsis.html"

    visual_placement = SideBlock(required=False, label="Visual Placement")

    show_publishing_date = blocks.BooleanBlock(required=False, label="Show Publishing Date")
    show_authors = blocks.BooleanBlock(required=False, label="Show Authors")
    show_rubrics = blocks.BooleanBlock(required=False, label="Show Rubrics")
    show_tags = blocks.BooleanBlock(required=False, label="Show Tags")

    visuals = SwitchBlock(local_blocks=[
        ('figures', FigureBlock(user_configurable_content=False)),
        ('slideshow', SlideshowBlock(user_configurable_content=False))],
        default_block_name='slideshow',
        template=APP_LABEL + "/blocks/visual_synopsis_switch.html")


class TextualSynopsisBlock(SynopsisStyleBlock):

    class Meta:
        template = APP_LABEL + "/blocks/textual_synopsis.html"

    show_publishing_date = blocks.BooleanBlock(required=False, label="Show Publishing Date")
    show_authors = blocks.BooleanBlock(required=False, label="Show Authors")
    show_rubrics = blocks.BooleanBlock(required=False, label="Show Rubrics")
    show_tags = blocks.BooleanBlock(required=False, label="Show Tags")


class SynopsisStyleBlockRegistry(BlockRegistry):

    # noinspection PyMethodMayBeStatic
    def should_include_entry_for_switch_block(self, identifier, entry, switch_block):
        return True

    # noinspection PyMethodMayBeStatic
    def instantiate_block(self, identifier, entry, switch_block):

        block_kwargs = dict(entry.block_kwargs)

        block_kwargs["user_configurable_content"] = False

        block = entry.block_type(*entry.block_args, **block_kwargs)
        block.set_name(identifier)

        return block


SYNOPSIS_STYLE_BLOCK_REGISTRY = SynopsisStyleBlockRegistry()
SYNOPSIS_STYLE_BLOCK_REGISTRY.define_procedures_in_caller_module("synopsis_style")

register_synopsis_style_block(APP_LABEL, "visual", VisualSynopsisBlock, [], {}) # noqa
register_synopsis_style_block(APP_LABEL, "textual", TextualSynopsisBlock, [], {}) # noqa


class SynopsesLayoutBlock(BaseContentLayoutBlock):

    class Meta:
        classname = settings.WAGTAIL_SYNOPSIS_SYNOPSES_LAYOUT_BLOCK_CLASSNAME
        tag_classname = settings.WAGTAIL_SYNOPSIS_TAG_CLASSNAME
        supported_item_types = [APP_LABEL + ".synopsis"]

    def deconstruct(self):
        return blocks.Block.deconstruct(self)


class SynopsisListBlock(SynopsesLayoutBlock):
    class Meta:
        classname = settings.WAGTAIL_SYNOPSIS_SYNOPSIS_LIST_BLOCK_CLASSNAME
        content_wrapper_classname = settings.WAGTAIL_SYNOPSIS_SYNOPSIS_LIST_CONTENT_WRAPPER_CLASSNAME
        template = APP_LABEL + "/blocks/synopsis_list.html"

    synopsis_style = DynamicSwitchBlock(child_blocks_function_name=APP_LABEL + ".blocks.synopsis_style_block_choices",
                                      choice_label="Select")

    def __init__(self, *args, **kwargs):

        self.base_blocks['synopsis_style'].container_block = weakref.ref(self)

        super().__init__(*args, **kwargs)





from wagtail.permission_policies import BlanketPermissionPolicy

from wagtail_content_admin.content_admin import ContentAdmin
from wagtail_richer_text.templatetags.wagtail_richer_text_tags import richertext

from .apps import get_app_label

__all__ = ['synopsis_admin', 'SynopsisAdmin']

APP_LABEL = get_app_label()


class SynopsisAdmin(ContentAdmin):

    browser_order_by = ['-created_at']
    url_namespace = APP_LABEL
    permission_policy = BlanketPermissionPolicy(APP_LABEL + ".synopsis")

    def __init__(self):
        super(SynopsisAdmin, self).__init__()

    def render_choice_inner(self, instance, **kwargs):
        result = richertext(instance.summary)
        return result

    def render_preview_inner(self, instance, **kwargs):
        result = richertext(instance.summary)


        """
          <h3 class="abstract-heading">
  {% if layout.template_values.show_rubrics %}{% if synopsis.rubric_url %}<a class="abstract-rubric" href="{{ synopsis.rubric_url }}">{{ synopsis.rubric }}</a>{% else %}{{ synopsis.rubric }}{% endif %}{% endif %}
  {% if synopsis.has_destination_url %}<a href="{% synopsis_destination_url synopsis request=request %}">{{ synopsis.heading }}</a>{% else %}{{ synopsis.heading }}{% endif %}</h3>
  {% if layout.template_values.show_publishing_date %}{% if synopsis.created_at or synopsis.updated_at %}
  <p class="abstract-dateline"><span class="abstract-publishing-date">{% if synopsis.updated_at %}{{ synopsis.updated_at }}{% else %}{{ synopsis.created_at }}{% endif %}</span></p>
  {% endif %}{% endif %}
          <div class="abstract-summary">{{ synopsis.summary|richertext }}</div>
        """

        return result


synopsis_admin = SynopsisAdmin()




from django.apps import apps


__all__ = ['register_synopsis_category', 'get_synopsis_category_choices', 'CuratedSynopsis']


SYNOPSIS_CATEGORIES = dict()


class SynopsisCategory:

    @property
    def identifier(self):
        return self.app_label + ":" + self.local_identifier

    def __init__(self, app_label, local_identifier, name):

        self.app_label = app_label
        self.local_identifier = local_identifier
        self.name = name

        app = apps.get_app_config(self.app_label)
        self.app_name = app.verbose_name


def register_synopsis_category(app_label, local_identifier, name):

    category = SynopsisCategory(app_label=app_label, local_identifier=local_identifier, name=name)
    SYNOPSIS_CATEGORIES[category.identifier] = category
    return category


def get_synopsis_category_choices():

    result = [(identifier, category.name + " [{}]".format(category.app_name.title()))
              for identifier, category in SYNOPSIS_CATEGORIES.items()]

    return result


class CuratedSynopsis(object):

    @property
    def synopsis(self):
        return self.synopsis_

    @property
    def prominence_css_class(self):
        return self.prominence_css_class_

    @prominence_css_class.setter
    def prominence_css_class(self, value):
        self.prominence_css_class_ = value

    @property
    def prominence_index(self):

        if self.prominence_index_ is None:
            return self.synopsis_.prominence_index

        return self.prominence_index_

    @prominence_index.setter
    def prominence_index(self, value):
        self.prominence_index_ = value

    @property
    def prominence_order(self):

        if self.prominence_order_ is None:
            return self.synopsis_.prominence_order

        return self.prominence_order_

    @prominence_order.setter
    def prominence_order(self, value):
        self.prominence_order_ = value

    def __getattr__(self, name):
        return getattr(self.synopsis_, name)

    def __init__(self, synopsis):
        self.synopsis_ = synopsis
        self.prominence_css_class_ = ''
        self.prominence_index_ = ''
        self.prominence_order_ = ''


# -*- coding: utf-8 -*-

from types import SimpleNamespace
from django.urls import reverse
from django.apps import AppConfig
from django.apps import apps
from django.utils.translation import gettext_lazy as _

from wagtail.utils.decorators import cached_classmethod

PAGE_EDIT_HANDLERS = set()


class WagtailSynopsisConfig(AppConfig):
    name = 'wagtail_synopsis'
    label = 'wagtail_synopsis'
    verbose_name = _("Wagtail Synopsis")
    default_auto_field = 'django.db.models.BigAutoField'
    app_settings_getters = SimpleNamespace()

    def import_models(self):

        from django_auxiliaries.app_settings import configure

        self.app_settings_getters = configure(self)

        super().import_models()

    def ready(self):

        from .models import PageSynopsis, sync_page_synopsis, determine_page_url
        from wagtail.models import Page

        page_adapter = PageSynopsis.install_synopsis_adapter_for(
            editable_fields=['summary', 'visual', 'tags'],
            sync_method=sync_page_synopsis,
            determine_url_method=determine_page_url)

        base_get_edit_handler = Page.get_edit_handler

        @cached_classmethod
        def get_page_edit_handler(cls):
            from wagtail.admin.panels import TabbedInterface

            # access the original function to derive the edit handler
            edit_handler = base_get_edit_handler.fn(cls) # noqa

            if isinstance(edit_handler, TabbedInterface) and id(edit_handler) not in PAGE_EDIT_HANDLERS:
                PAGE_EDIT_HANDLERS.add(id(edit_handler))
                synopsis_panel = page_adapter.create_synopsis_panel_for(cls)
                if synopsis_panel not in edit_handler.children:
                    edit_handler.children.insert(1, synopsis_panel)

            return edit_handler

        Page.get_edit_handler = get_page_edit_handler


def get_app_label():
    return WagtailSynopsisConfig.label


def reverse_app_url(identifier):
    return reverse(f'{WagtailSynopsisConfig.label}:{identifier}')


def get_app_config():
    return apps.get_app_config(WagtailSynopsisConfig.label)



from django.forms import fields
from django.forms import widgets

__all__ = ['StaticFormField']


class StaticFormField(fields.CharField):

    def __init__(self, *args, **kwargs):
        kwargs['widget'] = widgets.HiddenInput
        super().__init__(*args, **kwargs)

    def bound_data(self, data, initial):
        return ''

    def has_changed(self, initial, data):
        return True



from django import template
from django.templatetags.static import static

from ..apps import get_app_label

APP_LABEL = get_app_label()

register = template.Library()


@register.simple_tag(takes_context=False, name="synopsis_destination_url")
def synopsis_destination_url_tag(synopsis, request=None, site=None):
    return synopsis.determine_destination_url(request, site)


def gather_style_sheets():

    urls = [
            # static(APP_LABEL + '/css/synopsis_list.css')
    ]

    return urls


def gather_scripts():

    urls = []

    return urls


SUPPORT_TEMPLATE_SETTING = APP_LABEL + '/tags/support.html'


@register.inclusion_tag(SUPPORT_TEMPLATE_SETTING, name="wagtail_synopsis_support")
def wagtail_synopsis_support_tag(*, container_element, is_admin_page=False):

    result = {
        'container_element': container_element,
        'is_admin_page': is_admin_page,
        'stylesheets': gather_style_sheets(),
        'scripts': gather_scripts()
    }

    return result


from django.contrib.contenttypes.models import ContentType
from django.utils.functional import cached_property

__all__ = ['SpecificMixin']


class SpecificMixin:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not self.id: # noqa

            self.id = None

            if hasattr(self.__class__, 'content_type') and not self.content_type_id: # noqa

                self.content_type_id = ContentType.objects.get_for_model(self).id

    @property
    def cached_content_type(self):
        """
        .. versionadded:: 2.10

        Return this media_item's ``content_type`` value from the ``ContentType``
        model's cached manager, which will avoid a database query if the
        object is already in memory.
        """
        return ContentType.objects.get_for_id(
                self.content_type_id)  # noqa

    @cached_property
    def specific_class(self):
        """
        Return the class that this media item would be if instantiated in its
        most specific form.

        If the model class can no longer be found in the codebase, and the
        relevant ``ContentType`` has been removed by a database migration,
        the return value will be ``None``.

        If the model class can no longer be found in the codebase, but the
        relevant ``ContentType`` is still present in the database (usually a
        result of switching between git branches without running or reverting
        database migrations beforehand), the return value will be ``None``.
        """
        return self.cached_content_type.model_class()

    @cached_property
    def specific(self):
        """
        Returns this media_item in its most specific subclassed form with all field
        values fetched from the database. The result is cached in memory.
        """
        return self.get_specific()

    @cached_property
    def specific_deferred(self):
        """
        .. versionadded:: 2.12

        Returns this media_item in its most specific subclassed form without any
        additional field values being fetched from the database. The result
        is cached in memory.
        """
        return self.get_specific(deferred=True)

    def get_specific(self, deferred=False, copy_attrs=None):
        """
        .. versionadded:: 2.12

        Return this media_item in its most specific subclassed form.

        By default, a database query is made to fetch all field values for the
        specific object. If you only require access to custom methods or other
        non-field attributes on the specific object, you can use
        ``deferred=True`` to avoid this query. However, any attempts to access
        specific field values from the returned object will trigger additional
        database queries.

        If there are attribute values on this object that you wish to be copied
        over to the specific version (for example: evaluated relationship field
        values, annotations or cached properties), use `copy_attrs`` to pass an
        iterable of names of attributes you wish to be copied.

        If called on a media_item object that is already an instance of the most
        specific class (e.g. an ``EventPage``), the object will be returned
        as is, and no database queries or other operations will be triggered.

        If the media_item was originally created using a media_item type that has since
        been removed from the codebase, a generic ``Page`` object will be
        returned (without any custom field values or other functionality
        present on the original class). Usually, deleting these pages is the
        best course of action, but there is currently no safe way for Wagtail
        to do that at migration time.
        """

        model_class = self.specific_class

        if model_class is None:
            # The codebase and database are out of sync (e.g. the model exists
            # on a different git branch and migrations were not applied or
            # reverted before switching branches). So, the best we can do is
            # return the media_item in it's current form.
            return self

        if isinstance(self, model_class):
            # self is already the an instance of the most specific class
            return self

        if deferred:
            # Generate a tuple of values in the order expected by __init__(),
            # with missing values substituted with DEFERRED ()
            values = tuple(
                getattr(self, f.attname, self.pk if f.primary_key else DEFERRED)  # noqa
                for f in model_class._meta.concrete_fields
            )
            # Create object from known attribute values
            specific_obj = model_class(*values)
            specific_obj._state.adding = self._state.adding  # noqa
        else:
            # Fetch object from database
            specific_obj = model_class._default_manager.get(id=self.id)  # noqa

        # Copy additional attribute values
        for attr in copy_attrs or ():
            if attr in self.__dict__:
                setattr(specific_obj, attr, getattr(self, attr))

        return specific_obj




from django.urls import include, path
from wagtail import hooks

from . import admin_urls
from .apps import get_app_label

APP_LABEL = get_app_label()


@hooks.register('register_admin_urls')
def register_admin_urls():
    return [
        path(APP_LABEL + '/', include(admin_urls, namespace=APP_LABEL)),
    ]
